from __future__ import annotations

import logging
from collections.abc import Generator
from typing import cast, overload

import claripy
from claripy.annotation import RegionAnnotation, StridedIntervalAnnotation
from claripy.ast import BV, Base, Bool
from claripy.errors import BackendError, ClaripyOperationError

from angr.errors import AngrError

log = logging.getLogger(__name__)

commutative_operations = {
    "__and__",
    "__or__",
    "__xor__",
    "__add__",
    "__mul__",
    "And",
    "Or",
    "Xor",
}


class BalancerError(AngrError):
    """
    Base class for balancer errors.
    """


class BalancerUnsatError(BalancerError):
    """
    Exception raised when the balancer determines the constraints are unsatisfiable.
    """


class Balancer:
    """
    The Balancer is an equation redistributor. The idea is to take an AST and rebalance it to, for example, isolate
    unknown terms on one side of an inequality.
    """

    def __init__(self, c: Bool):
        self._truisms: list[Bool] = []
        self._ast_hash_map: dict[int, BV] = {}
        self._lower_bounds: dict[int, int] = {}
        self._upper_bounds: dict[int, int] = {}

        self.sat = True
        try:
            self._doit(c)
        except BalancerUnsatError:
            self.bounds = {}
            self.sat = False
        except BackendError:
            log.debug("Backend error in balancer.", exc_info=True)

    @property
    def compat_ret(self) -> tuple[bool, list[tuple[BV, BV]]]:
        return (self.sat, self.replacements)

    def _replacements_iter(self) -> Generator[tuple[BV, BV]]:
        all_keys = set(self._lower_bounds.keys()) | set(self._upper_bounds.keys())
        for k in all_keys:
            ast = self._ast_hash_map[k]
            max_int = (1 << len(ast)) - 1
            min_int = 0
            mn = self._lower_bounds.get(k, min_int)
            mx = self._upper_bounds.get(k, max_int)
            bound_si = claripy.BVS("bound", len(ast)).annotate(StridedIntervalAnnotation(1, mn, mx))
            log.debug("Yielding bound %s for %s.", bound_si, ast)
            if ast.op == "Reverse":
                yield (cast(BV, ast.args[0]), ast.intersection(bound_si).reversed)
            else:
                yield (ast, ast.intersection(bound_si))

    def _add_lower_bound(self, o: BV, b: int) -> None:
        if o.hash() in self._lower_bounds:
            old_b = self._lower_bounds[o.hash()]
            b = max(b, old_b)

        self._lower_bounds[o.hash()] = b
        self._ast_hash_map[o.hash()] = o

    def _add_upper_bound(self, o: BV, b: int) -> None:
        if o.hash() in self._upper_bounds:
            old_b = self._upper_bounds[o.hash()]
            b = min(b, old_b)

        self._upper_bounds[o.hash()] = b
        self._ast_hash_map[o.hash()] = o

    @property
    def replacements(self) -> list[tuple[BV, BV]]:
        return list(self._replacements_iter())

    #
    # AST helper functions
    #

    @staticmethod
    def _same_bound_bv(a: BV) -> BV:
        si = claripy.backends.vsa.simplify(a)
        mx = Balancer._max(a)
        mn = Balancer._min(a)
        si_anno = si.get_annotation(StridedIntervalAnnotation)
        stride = si_anno.stride if si_anno is not None else 0
        return claripy.BVS("bounds", len(a)).annotate(StridedIntervalAnnotation(stride, mn, mx))

    @staticmethod
    def _min(a: BV, signed=False) -> int:
        a = claripy.backends.vsa.simplify(a)
        if a.has_annotation_type(RegionAnnotation):
            region_annos = a.get_annotations_by_type(RegionAnnotation)
            if len(region_annos) == 1:
                a = next(iter(region_annos)).region_base_addr
            else:
                # unfortunately, this is a real abstract pointer
                # the minimum value will be 0 or MIN_INT
                if signed:
                    return -(1 << (len(a) - 1))
                return 0
        return claripy.backends.vsa.min(a, signed=signed)

    @staticmethod
    def _max(a: BV, signed=False) -> int:
        a = claripy.backends.vsa.simplify(a)
        if a.has_annotation_type(RegionAnnotation):
            region_annos = a.get_annotations_by_type(RegionAnnotation)
            if len(region_annos) == 1:
                a = next(iter(region_annos)).region_base_addr
            else:
                # unfortunately, this is a real abstract pointer
                # the minimum value will be 0 or MIN_INT
                if signed:
                    return (1 << (len(a) - 1)) - 1
                return (1 << len(a)) - 1
        return claripy.backends.vsa.max(a, signed=signed)

    @staticmethod
    def _range(a: BV, signed=False) -> tuple[int, int]:
        return (Balancer._min(a, signed=signed), Balancer._max(a, signed=signed))

    #
    # Truism alignment
    #

    @staticmethod
    def _align_truism(truism: Bool) -> Bool:
        outer_aligned = Balancer._align_ast(truism)
        inner_aligned = Bool(
            outer_aligned.op, (Balancer._align_ast(cast(BV, outer_aligned.args[0])), *outer_aligned.args[1:])
        )

        if not claripy.backends.vsa.identical(inner_aligned, truism):
            log.critical(
                "ERROR: the balancer is messing up an AST. This must be looked into. "
                "Please submit the binary and script to the angr project, if possible. "
                "Outer op is %s and inner op is %s.",
                truism.op,
                cast(BV, truism.args[0]).op,
            )
            return truism

        return cast(Bool, inner_aligned)

    @overload
    @staticmethod
    def _align_ast(a: Bool) -> Bool: ...

    @overload
    @staticmethod
    def _align_ast(a: BV) -> BV: ...

    @staticmethod
    def _align_ast(a: Bool | BV) -> Bool | BV:
        """
        Aligns the AST so that the argument with the highest cardinality is on the left.

        :return: a new AST.
        """

        try:
            if isinstance(a, BV):
                return Balancer._align_bv(a)
            if (
                isinstance(a, Bool)
                and len(a.args) == 2
                and cast(Bool, a.args[1]).cardinality > cast(Bool, a.args[0]).cardinality
            ):
                return Balancer._reverse_comparison(a)
            return a
        except BalancerError:
            return a

    @staticmethod
    def _reverse_comparison(a: Bool) -> Bool:
        lhs, rhs = cast(tuple[BV, BV], a.args)
        match a.op:
            case "__eq__":
                return rhs == lhs
            case "__ne__":
                return rhs != lhs
            case "ULT":
                return claripy.UGT(rhs, lhs)
            case "ULE":
                return claripy.UGE(rhs, lhs)
            case "UGT":
                return claripy.ULT(rhs, lhs)
            case "UGE":
                return claripy.ULE(rhs, lhs)
            case "SLT":
                return claripy.SGT(rhs, lhs)
            case "SLE":
                return claripy.SGE(rhs, lhs)
            case "SGT":
                return claripy.SLT(rhs, lhs)
            case "SGE":
                return claripy.SLE(rhs, lhs)
            case _:
                raise BalancerError(f"unable to reverse comparison {a.op}")

    @staticmethod
    def _align_bv(a: BV) -> BV:
        if a.op in commutative_operations:
            return BV(a.op, tuple(sorted(cast(tuple[BV, ...], a.args), key=lambda v: -v.cardinality)), length=a.length)

        match a.op:
            case "__sub__":
                return Balancer._align_sub(a)
            case _:
                return a

    @staticmethod
    def _align_sub(a: BV) -> BV:
        args = cast(tuple[BV, ...], a.args)

        cardinalities = [v.cardinality for v in args]
        if max(cardinalities) == cardinalities[0]:
            return a
        adjusted = tuple(-v for v in args[1:]) + args[:1]
        return BV("__add__", tuple(sorted(adjusted, key=lambda v: -v.cardinality)), length=a.length)

    #
    # Find bounds
    #

    def _doit(self, c: Bool) -> None:
        """
        This function processes the list of truisms and finds bounds for ASTs.
        """
        self._truisms.append(claripy.excavate_ite(c))

        processed_truisms = set()
        identified_assumptions = set()

        while len(self._truisms):
            truism = self._truisms.pop()

            if truism in processed_truisms:
                continue

            unpacked_truisms = Balancer._unpack_truisms(truism)
            if claripy.backends.vsa.is_false(truism):
                raise BalancerUnsatError

            processed_truisms.add(truism)
            if len(unpacked_truisms):
                self._truisms.extend(t for t in unpacked_truisms if not claripy.backends.vsa.is_true(t))
                continue

            if not Balancer._handleable_truism(truism):
                continue

            truism = Balancer._adjust_truism(truism)

            assumptions = Balancer._get_assumptions(truism)
            if truism not in identified_assumptions and len(assumptions):
                log.debug("Queued assumptions %s for truism %s.", assumptions, truism)
                self._truisms.extend(assumptions)
                identified_assumptions.update(assumptions)

            log.debug("Processing truism %s", truism)
            balanced_truism = self._balance(truism)
            log.debug("... handling")
            self._handle(balanced_truism)

    @staticmethod
    def _handleable_truism(t: Bool) -> bool:
        """
        Checks whether we can handle this truism. The truism should already be aligned.
        """
        if len(t.args) < 2:
            log.debug("can't do anything with an unop bool")
            return False
        lhs = cast(BV, t.args[0])
        rhs = cast(BV, t.args[1])
        if lhs.cardinality > 1 and rhs.cardinality > 1:
            log.debug("can't do anything because we have multiple multivalued guys")
            return False
        if t.op == "If":
            log.debug("can't handle If")
            return False
        return True

    @staticmethod
    def _adjust_truism(t: Bool) -> Bool:
        """
        Swap the operands of the truism if the unknown variable is on the right side and the concrete value is on the
        left side.
        """
        lhs, rhs = cast(tuple[BV, BV], t.args)
        if lhs.cardinality == 1 and rhs.cardinality > 1:
            return Balancer._reverse_comparison(t)
        return t

    #
    # Assumptions management
    #

    @staticmethod
    def _get_assumptions(t: Bool) -> list[Bool]:
        """
        Given a constraint, _get_assumptions() returns a set of constraints that are implicitly
        assumed to be true. For example, `x <= 10` would return `x >= 0`.
        """
        lhs = cast(BV, t.args[0])

        if t.op in ("ULE", "ULT"):
            return [lhs >= 0]
        if t.op in ("UGE", "UGT"):
            return [lhs <= 2 ** len(lhs) - 1]
        if t.op in ("SLE", "SLT"):
            return [claripy.SGE(lhs, -(1 << (len(lhs) - 1)))]
        if t.op in ("SGE", "SGT"):
            return [claripy.SLE(lhs, (1 << (len(lhs) - 1)) - 1)]
        return []

    #
    # Truism extractor
    #

    @staticmethod
    def _unpack_truisms(c: Bool) -> set[Bool]:
        """
        Given a constraint, _unpack_truisms() returns a set of constraints that must be True for
        this constraint to be True.
        """
        match c.op:
            case "And":
                return Balancer._unpack_truisms_and(c)
            case "Not":
                return Balancer._unpack_truisms_not(c)
            case "Or":
                return Balancer._unpack_truisms_or(c)
            case _:
                return set()

    @staticmethod
    def _unpack_truisms_and(c: Bool) -> set[Bool]:
        return set.union(*[Balancer._unpack_truisms(a) for a in cast(tuple[Bool, ...], c.args)])

    @staticmethod
    def _unpack_truisms_not(c: Bool) -> set[Bool]:
        arg = cast(Bool, c.args[0])
        if arg.op == "And":
            return Balancer._unpack_truisms(claripy.Or(*[claripy.Not(a) for a in arg.args]))
        if arg.op == "Or":
            return Balancer._unpack_truisms(claripy.And(*[claripy.Not(a) for a in arg.args]))
        return set()

    @staticmethod
    def _unpack_truisms_or(c: Bool) -> set[Bool]:
        args = cast(tuple[Bool, ...], c.args)
        vals = [claripy.backends.vsa.is_false(v) for v in args]
        if all(vals):
            raise BalancerUnsatError
        if vals.count(False) == 1:
            return Balancer._unpack_truisms(args[vals.index(False)])
        return set()

    #
    # Simplification routines
    #

    def _balance(self, truism: Bool) -> Bool:
        while True:
            log.debug("Balancing %s", truism)

            # can't balance single-arg bools (Not) for now
            if len(truism.args) == 1:
                return truism

            if not isinstance(truism.args[0], Base):
                return truism

            try:
                inner_aligned = Balancer._align_truism(truism)
                inner_lhs = cast(BV, inner_aligned.args[0])
                inner_rhs = cast(BV, inner_aligned.args[1])
                if inner_rhs.cardinality > 1:
                    log.debug("can't do anything because we have multiple multivalued guys")
                    return truism

                match inner_lhs.op:
                    case "Reverse":
                        balanced = Balancer._balance_reverse(inner_aligned)
                    case "__add__":
                        balanced = Balancer._balance_add(inner_aligned)
                    case "__sub__":
                        balanced = Balancer._balance_sub(inner_aligned)
                    case "ZeroExt":
                        balanced = Balancer._balance_zeroext(inner_aligned)
                    case "SignExt":
                        balanced = Balancer._balance_signext(inner_aligned)
                    case "Extract":
                        balanced = Balancer._balance_extract(inner_aligned)
                    case "__and__":
                        balanced = Balancer._balance_and(inner_aligned)
                    case "Concat":
                        balanced = Balancer._balance_concat(inner_aligned)
                    case "__lshift__":
                        balanced = Balancer._balance_lshift(inner_aligned)
                    case "If":
                        balanced = self._balance_if(inner_aligned)
                    case _:
                        log.debug("Balance handler %s not implemented.", truism.args[0].op)
                        return truism

                if balanced is inner_aligned:
                    return balanced
                truism = balanced
                continue
            except BalancerError:
                log.warning("Balance handler for operation %s raised exception.", truism.args[0].op)
                return truism

    @staticmethod
    def _balance_reverse(truism: Bool) -> Bool:
        if truism.op in ["__eq__", "__ne__"]:
            return BV(
                truism.op,
                (cast(BV, truism.args[0]).args[0], cast(BV, truism.args[1]).reversed),
                length=cast(BV, truism.args[1]).length,
            )
        return truism

    @staticmethod
    def _balance_add(truism: Bool) -> Bool:
        if len(truism.args) != 2:
            return truism
        lhs, old_rhs = cast(tuple[BV, BV], truism.args)
        lhs_args = cast(tuple[BV, ...], lhs.args)
        if all(a.concrete for a in lhs_args):
            # the old logic
            new_lhs = lhs_args[0]
            other_adds = lhs_args[1:]
        else:
            new_lhs = tuple(a for a in lhs_args if a.symbolic)
            if not new_lhs:
                return truism
            new_lhs = new_lhs[0] if len(new_lhs) == 1 else BV("__add__", new_lhs, length=lhs.length)
            other_adds = tuple(a for a in lhs_args if a.concrete)
            if not other_adds:
                return truism
        new_rhs = BV("__sub__", (old_rhs, *other_adds), length=lhs.length)
        return Bool(truism.op, (new_lhs, new_rhs))

    @staticmethod
    def _balance_sub(truism: Bool) -> Bool:
        if len(truism.args) != 2:
            return truism
        old_lhs = cast(BV, truism.args[0])
        new_lhs = cast(BV, old_lhs.args[0])
        old_rhs = cast(BV, truism.args[1])
        other_adds = old_lhs.args[1:]
        new_rhs = BV("__add__", (old_rhs, *other_adds), length=old_lhs.length)
        return Bool(truism.op, (new_lhs, new_rhs))

    @staticmethod
    def _balance_zeroext(truism: Bool) -> Bool:
        orig_lhs = cast(BV, truism.args[0])
        orig_rhs = cast(BV, truism.args[1])

        num_zeroes, inner = cast(tuple[int, BV], orig_lhs.args)
        other_side = orig_rhs[len(orig_rhs) - 1 : len(orig_rhs) - num_zeroes]

        if claripy.backends.vsa.is_true(other_side == 0):
            # We can safely eliminate this layer of ZeroExt
            new_args = (inner, orig_rhs[len(orig_rhs) - num_zeroes - 1 : 0])
            return Bool(truism.op, new_args)

        return truism

    @staticmethod
    def _balance_signext(truism: Bool) -> Bool:
        orig_lhs = cast(BV, truism.args[0])
        orig_rhs = cast(BV, truism.args[1])

        num_zeroes = cast(int, orig_lhs.args[0])
        left_side = orig_lhs[len(orig_rhs) - 1 : len(orig_rhs) - num_zeroes]
        other_side = orig_rhs[len(orig_rhs) - 1 : len(orig_rhs) - num_zeroes]

        # TODO: what if this is a set value, but *not* the same as other_side
        if claripy.backends.vsa.identical(left_side, other_side):
            # We can safely eliminate this layer of ZeroExt
            new_args = (orig_lhs.args[1], orig_rhs[len(orig_rhs) - num_zeroes - 1 : 0])
            return Bool(truism.op, new_args)

        return truism

    @staticmethod
    def _balance_extract(truism: Bool) -> Bool:
        lhs = cast(BV, truism.args[0])
        high, low, inner = cast(tuple[int, int, BV], lhs.args)
        inner_size = len(inner)
        rhs = cast(BV, truism.args[1])

        left_msb: BV | None
        left_msb_zero: bool | None

        if high < inner_size - 1:
            left_msb = inner[inner_size - 1 : high + 1]
            left_msb_zero = claripy.backends.vsa.is_true(left_msb == 0)
        else:
            left_msb = None
            left_msb_zero = None

        left_lsb: BV | None
        left_lsb_zero: bool | None

        if low > 0:
            left_lsb = inner[high - 1 : 0]
            left_lsb_zero = claripy.backends.vsa.is_true(left_lsb == 0)
        else:
            left_lsb = None
            left_lsb_zero = None

        if left_msb_zero and left_lsb_zero:
            new_left = inner
            new_right = claripy.Concat(
                claripy.BVV(0, len(cast(BV, left_msb))), rhs, claripy.BVV(0, len(cast(BV, left_lsb)))
            )
            return Bool(truism.op, (new_left, new_right))
        if left_msb_zero:
            new_left = inner
            new_right = claripy.Concat(claripy.BVV(0, len(cast(BV, left_msb))), rhs)
            return Bool(truism.op, (new_left, new_right))
        if left_lsb_zero:
            new_left = inner
            new_right = claripy.Concat(rhs, claripy.BVV(0, len(cast(BV, left_lsb))))
            return Bool(truism.op, (new_left, new_right))
        if low == 0 and rhs.op == "BVV" and truism.op not in {"SGE", "SLE", "SGT", "SLT"}:
            # single-valued rhs value with an unsigned operator
            # Eliminate Extract on lhs and zero-extend the value on rhs
            new_left = inner
            new_right = claripy.ZeroExt(inner.size() - rhs.size(), rhs)
            return Bool(truism.op, (new_left, new_right))

        return truism

    @staticmethod
    def _balance_and(truism: Bool) -> Bool:
        lhs = cast(BV, truism.args[0])
        if len(lhs.args) != 2:
            return truism
        op0, op1 = cast(tuple[BV, BV], lhs.args)

        if op1.op == "BVV":
            # if all low bits of right are 1 and all high bits of right are 0, then this is equivalent to Extract()
            v = cast(int, op1.args[0])
            low_ones = 0
            while v != 0:
                if v & 1 == 0:
                    # not all high bits are 0. abort
                    return truism
                low_ones += 1
                v >>= 1
            if low_ones == 0:
                # this should probably never happen
                new_left = BV("BVV", (0, op0.size()), length=op0.size())
                return Bool(truism.op, (new_left, truism.args[1]))

            if op0.op == "ZeroExt" and cast(int, op0.args[0]) + low_ones == op0.size():
                # ZeroExt(56, a) & 0xff == a  if a.size() == 8
                # we can safely remove __and__
                new_left = op0
                return Bool(truism.op, (new_left, truism.args[1]))

        return truism

    @staticmethod
    def _balance_concat(truism: Bool) -> Bool:
        lhs = cast(BV, truism.args[0])
        rhs = cast(BV, truism.args[1])
        size = len(lhs)
        left_msb = cast(BV, lhs.args[0])
        right_msb = rhs[size - 1 : size - len(left_msb)]

        if claripy.backends.vsa.is_true(left_msb == 0) and claripy.backends.vsa.is_true(right_msb == 0):
            # we can cut these guys off!
            lhs_args = cast(tuple[BV, ...], lhs.args)
            remaining_left = claripy.Concat(*lhs_args[1:])
            remaining_right = rhs[size - len(left_msb) - 1 : 0]
            return Bool(truism.op, (remaining_left, remaining_right))
        # TODO: handle non-zero single-valued cases
        return truism

    @staticmethod
    def _balance_lshift(truism: Bool) -> Bool:
        lhs = cast(BV, truism.args[0])
        rhs = cast(BV, truism.args[1])
        shift_amount_expr = cast(BV, lhs.args[1])
        expr = cast(BV, lhs.args[0])

        shift_amount_values = claripy.backends.vsa.eval(shift_amount_expr, 2)
        if len(shift_amount_values) != 1:
            return truism
        shift_amount = shift_amount_values[0]

        rhs_lower = claripy.Extract(shift_amount - 1, 0, rhs)
        rhs_lower_values = claripy.backends.vsa.eval(rhs_lower, 2)
        if len(rhs_lower_values) == 1 and rhs_lower_values[0] == 0:
            # we can remove the __lshift__

            return Bool(truism.op, (expr, rhs >> shift_amount))

        return truism

    def _balance_if(self, truism: Bool) -> Bool:
        lhs = cast(BV, truism.args[0])
        condition, true_expr, false_expr = cast(tuple[Bool, BV, BV], lhs.args)

        try:
            true_condition = getattr(true_expr, truism.op)(truism.args[1])
            false_condition = getattr(false_expr, truism.op)(truism.args[1])
        except ClaripyOperationError:
            # the condition was probably a Not (TODO)
            return truism

        can_true = claripy.backends.vsa.has_true(true_condition)
        can_false = claripy.backends.vsa.has_true(false_condition)
        must_true = claripy.backends.vsa.is_true(true_condition)
        must_false = claripy.backends.vsa.is_true(false_condition)

        if can_true and can_false:
            # always satisfiable
            return truism
        if not (can_true or can_false):
            # neither are satisfiable. This truism is fucked
            raise BalancerUnsatError
        if must_true or (can_true and not can_false):
            # it will always be true
            self._truisms.append(condition)
            return Bool(truism.op, (true_expr, truism.args[1]))
        if must_false or (can_false and not can_true):
            # it will always be false
            self._truisms.append(~condition)
            return Bool(truism.op, (false_expr, truism.args[1]))
        raise BalancerError("unhandled If balancing case")

    #
    # Constraint handlers
    #

    def _handle(self, truism: Bool) -> None:
        log.debug("Handling %s", truism)

        if claripy.backends.vsa.is_false(truism):
            raise BalancerUnsatError
        if truism.op in {"BoolV", "BoolS"} or cast(Base, truism.args[0]).cardinality == 1:
            # we are down to single-cardinality arguments, so our work is not
            # necessary
            return

        match truism.op:
            case "__eq__":
                self._handle_eq(truism)
            case "__ne__":
                self._handle_ne(truism)
            case "If":
                self._handle_if(truism)
            case "ULT" | "ULE" | "UGT" | "UGE" | "SLT" | "SLE" | "SGT" | "SGE":
                self._handle_comparison(truism)
            case _:
                log.debug("No handler for operation %s", truism.op)

    comparison_info = {
        "ULT": (True, False, True),
        "ULE": (True, True, True),
        "UGT": (False, False, True),
        "UGE": (False, True, True),
        "SLT": (True, False, False),
        "SLE": (True, True, False),
        "SGT": (False, False, False),
        "SGE": (False, True, False),
    }

    def _handle_comparison(self, truism: Bool) -> None:
        """
        Handles all comparisons.
        """

        is_lt, is_equal, is_unsigned = self.comparison_info[truism.op]
        lhs, rhs = cast(tuple[BV, BV], truism.args)

        size = len(lhs)
        int_max = 2**size - 1 if is_unsigned else 2 ** (size - 1) - 1
        int_min = -(2 ** (size - 1))

        left_min = Balancer._min(lhs, signed=not is_unsigned)
        left_max = Balancer._max(lhs, signed=not is_unsigned)
        right_min = Balancer._min(rhs, signed=not is_unsigned)
        right_max = Balancer._max(rhs, signed=not is_unsigned)

        bound_max = right_max if is_equal else (right_max - 1 if is_lt else right_max + 1)
        bound_min = right_min if is_equal else (right_min - 1 if is_lt else right_min + 1)

        if is_lt and bound_max < int_min:
            # if the bound max is negative and we're unsigned less than, we're fucked
            raise BalancerUnsatError
        if not is_lt and bound_min > int_max:
            # if the bound min is too big, we're fucked
            raise BalancerUnsatError

        current_min = int_min
        current_max = int_max

        if is_lt:
            current_max = min(int_max, left_max, bound_max)
            self._add_upper_bound(lhs, current_max)
        else:
            current_min = max(int_min, left_min, bound_min)
            self._add_lower_bound(lhs, current_min)

    def _handle_eq(self, truism: Bool) -> None:
        lhs, rhs = cast(tuple[BV, BV], truism.args)
        if rhs.cardinality != 1:
            common = Balancer._same_bound_bv(lhs.intersection(rhs))
            mn, mx = Balancer._range(common)
            self._add_upper_bound(lhs, mx)
            self._add_upper_bound(rhs, mx)
            self._add_lower_bound(lhs, mn)
            self._add_lower_bound(rhs, mn)
        else:
            mn, mx = Balancer._range(rhs)
            self._add_upper_bound(lhs, mx)
            self._add_lower_bound(lhs, mn)

    def _handle_ne(self, truism: Bool) -> None:
        lhs, rhs = cast(tuple[BV, BV], truism.args)
        if rhs.cardinality == 1:
            val = claripy.backends.vsa.eval(rhs, 1)[0]
            max_int = claripy.BVV((1 << len(lhs)) - 1, len(lhs)).args[0]

            if val == 0:
                self._add_lower_bound(lhs, val + 1)
            elif val in (max_int, val - 1):
                self._add_upper_bound(lhs, max_int - 1)

    def _handle_if(self, truism: Bool) -> None:
        condition, true_expr, false_expr = cast(tuple[Bool, BV, BV], truism.args)
        if claripy.backends.vsa.is_false(false_expr):
            self._truisms.append(condition)
        elif claripy.backends.vsa.is_false(true_expr):
            self._truisms.append(claripy.Not(condition))


def constraint_to_si(expr: Bool) -> tuple[bool, list[tuple[BV, BV]]]:
    """
    Convert a constraint to SI if possible.

    :param expr:
    :return:
    """

    return Balancer(expr).compat_ret



from __future__ import annotations
import angr


class InterlockedExchange(angr.SimProcedure):
    def run(self, target, value):  # pylint:disable=arguments-differ
        if not self.state.solver.symbolic(target):
            old_value = self.state.memory.load(target, 4, endness=self.state.arch.memory_endness)
            self.state.memory.store(target, value)
        else:
            old_value = self.state.solver.Unconstrained(
                f"unconstrained_ret_{self.display_name}", self.state.arch.bits, key=("api", "InterlockedExchange")
            )

        return old_value



from __future__ import annotations

# pylint: disable=import-outside-toplevel
from typing import TYPE_CHECKING
from collections.abc import Callable, Iterable

import claripy

from angr.engines.ail.callstack import AILCallStack
from angr.sim_state import SimState
from angr.storage.memory_mixins.memory_mixin import MemoryMixin

if TYPE_CHECKING:
    from angr.analyses.decompiler.clinic import Clinic
    from angr.project import Project
    from angr.ailment import Address


def ail_call_state(
    project: Project,
    start_addr: int | Address | str,
    args: Iterable[claripy.ast.Bits],
    lifter: Callable[[int], Clinic],
    mode: str = "symbolic",
    options: set[str] | None = None,
    add_options: set[str] | None = None,
    remove_options: set[str] | None = None,
    plugin_preset: str = "default",
    memory_cls: type[MemoryMixin] | None = None,
):
    state = SimState(
        project,
        project.arch,
        mode=mode,
        add_options=add_options,
        remove_options=remove_options,
        options=options,
        cle_memory_backer=project.loader.memory,
        plugin_preset=plugin_preset,
    )
    # break circular imports. this module should maybe live somewhere else
    from angr.storage import DefaultMemory

    state.globals["ail_var_memory_cls"] = memory_cls or DefaultMemory  # type: ignore
    state.globals["ail_lifter"] = lifter  # type: ignore

    if isinstance(start_addr, str):
        start_addr = project.kb.functions[start_addr].addr
    state.addr = (start_addr, None) if isinstance(start_addr, int) else start_addr

    bottom_frame = AILCallStack()
    top_frame = AILCallStack(func_addr=start_addr)
    top_frame.passed_args = tuple(args)
    state.register_plugin("callstack", bottom_frame)
    state.callstack.push(top_frame)

    return state



# pylint:disable=consider-using-dict-items
from __future__ import annotations
from typing import Any
from collections import defaultdict

import networkx

from angr.utils.graph import shallow_reverse


class IncrementalDominators:
    """
    This class allows for incrementally updating dominators and post-dominators for graphs. The graph must only be
    modified by replacing nodes, not adding nodes or edges.
    """

    def __init__(self, graph: networkx.DiGraph, start, post: bool = False):
        self.graph = graph
        self.start = start
        self._post: bool = post  # calculate post-dominators if True
        self._pre: bool = not post  # calculate dominators

        self._doms: dict[Any, Any] = {}
        self._dfs: dict[Any, set[Any]] | None = None  # initialized on-demand
        self._inverted_dom_tree: dict[Any, Any] | None = None  # initialized on demand

        self._doms = self.init_doms()

    def init_doms(self) -> dict[Any, Any]:
        if self._post:
            t = shallow_reverse(self.graph)
            doms = networkx.immediate_dominators(t, self.start)
        else:
            doms = networkx.immediate_dominators(self.graph, self.start)
        doms[self.start] = self.start
        return doms

    def init_dfs(self) -> dict[Any, set[Any]]:
        _pred = self.graph.predecessors if self._pre else self.graph.successors
        df: dict = {}
        for u in self._doms:
            _preds = list(_pred(u))  # type: ignore
            if len(_preds) >= 2:
                for v in _preds:
                    if v in self._doms:
                        while v is not self._doms[u]:
                            if v not in df:
                                df[v] = set()
                            df[v].add(u)
                            v = self._doms[v]
        return df

    def _update_inverted_domtree(self):
        # recalculate the dominators for dominatees of replaced nodes
        if self._inverted_dom_tree is None:
            self._inverted_dom_tree = defaultdict(list)
            for dtee, dtor in self._doms.items():
                self._inverted_dom_tree[dtor].append(dtee)

    def graph_updated(self, new_node: Any, replaced_nodes: set[Any], replaced_head: Any):
        self._update_inverted_domtree()
        assert self._inverted_dom_tree is not None

        # recalculate the dominators for impacted nodes
        new_dom = self._doms[replaced_head]
        while new_dom in replaced_nodes and new_dom is not self.start:
            new_dom = self._doms[new_dom]

        if self.start in replaced_nodes:
            self.start = new_node
        if new_dom in replaced_nodes:
            new_dom = new_node

        new_node_doms = []
        for rn in replaced_nodes:
            if rn not in self._inverted_dom_tree:
                continue
            for dtee in self._inverted_dom_tree[rn]:
                self._doms[dtee] = new_node
                new_node_doms.append(dtee)
        self._doms[new_node] = new_dom

        if self._dfs is not None:
            # update dominance frontiers
            if replaced_head in self._dfs:
                self._dfs[new_node] = self._dfs[replaced_head]
            for rn in replaced_nodes:
                if rn in self._dfs:
                    del self._dfs[rn]
                for df in self._dfs.values():
                    if rn in df:
                        df.remove(rn)
                        df.add(new_node)

        # keep inverted dom tree up-to-date
        self._inverted_dom_tree[new_dom].append(new_node)
        self._inverted_dom_tree[new_node] = new_node_doms
        for rn in replaced_nodes:
            if rn in self._doms:
                d = self._doms[rn]
                del self._doms[rn]
                self._inverted_dom_tree[d].remove(rn)
            if rn in self._inverted_dom_tree:
                del self._inverted_dom_tree[rn]

    def idom(self, node: Any) -> Any | None:
        """
        Get the immediate dominator of a given node.
        """

        return self._doms.get(node, None)

    def df(self, node: Any) -> set[Any]:
        """
        Generate the dominance frontier of a node.
        """
        if self._dfs is None:
            self._dfs = self.init_dfs()
        return self._dfs.get(node, set())

    def dominates(self, dominator_node: Any, node: Any) -> bool:
        """
        Tests if dominator_node dominates (or post-dominates) node.
        """

        n = node
        while n:
            if n is dominator_node:
                return True
            d = self.idom(n)
            n = d if d is not None and n is not d else None
        return False

    def _debug_check(self):
        true_doms = self.init_doms()
        if len(true_doms) != len(self._doms):
            raise ValueError("dominators do not match")
        for k in true_doms:
            if true_doms[k] != self._doms[k]:
                print(f"{k!r}: {true_doms[k]!r} {self._doms[k]!r}")
                raise ValueError("dominators do not match")

        if self._dfs is not None:
            dfs = self.init_dfs()
            if len(dfs) != len(self._dfs):
                raise ValueError("dfs do not match")
            for k in dfs:
                if dfs[k] != self._dfs[k]:
                    print(f"{k!r}: {dfs[k]!r} {self._dfs[k]!r}")
                    raise ValueError("dfs do not match")



from __future__ import annotations
import claripy

import angr


class isxdigit(angr.SimProcedure):
    # pylint: disable=arguments-differ, missing-class-docstring
    def run(self, c):
        is_digit = claripy.And(c >= 48, c <= 57)
        is_upper_hex = claripy.And(c >= 65, c <= 70)
        is_lower_hex = claripy.And(c >= 97, c <= 102)
        int_size = self.arch.sizeof["int"]  # type: ignore[reportOptionalMemberAccess]

        return claripy.If(
            claripy.Or(is_digit, is_upper_hex, is_lower_hex),
            claripy.BVV(1, int_size),
            claripy.BVV(0, int_size),
        )



from __future__ import annotations
from typing import Any, Literal, overload
from collections import defaultdict
from collections.abc import Callable, Iterable

import networkx

import archinfo
from angr.ailment import Expression, Block, Address
from angr.ailment.expression import (
    Convert,
    Extract,
    Insert,
    VirtualVariable,
    Const,
    Phi,
    Tmp,
    Load,
    Register,
    StackBaseOffset,
    DirtyExpression,
    ITE,
    UnaryOp,
)
from angr.ailment.expression import Call
from angr.ailment.statement import Statement, Assignment, Store, CAS, SideEffectStatement
from angr.ailment.block_walker import AILBlockViewer

from angr.knowledge_plugins.key_definitions import atoms
from angr.code_location import AILCodeLocation
from .vvar_uses_collector import VVarUsesCollector
from .tmp_uses_collector import TmpUsesCollector
from .vvar_extra_defs_collector import FindExtraDefs

DEPHI_VVAR_REG_OFFSET = 4096


@overload
def get_reg_offset_base_and_size(
    reg_offset: int, arch: archinfo.Arch, size: int | None = None, resilient: Literal[True] = True
) -> tuple[int, int]: ...
@overload
def get_reg_offset_base_and_size(
    reg_offset: int, arch: archinfo.Arch, size: int | None = None, resilient: Literal[False] = False
) -> tuple[int, int] | None: ...


def get_reg_offset_base_and_size(
    reg_offset: int, arch: archinfo.Arch, size: int | None = None, resilient: bool = True
) -> tuple[int, int | None] | None:
    """
    Translate a given register offset into the offset of its full register and obtain the size of the full register.

    :param reg_offset:  The offset of the register to translate.
    :param arch:        The corresponding Arch object.
    :param size:        Size of the register to translate. Optional.
    :param resilient:   When set to True, this function will return the provided offset and size for registers that the
                        arch does not know about.
    :return:            A tuple of translated offset and the size of the full register.
    """

    base_reg_and_size = arch.get_base_register(reg_offset, size=size)
    if resilient and base_reg_and_size is None:
        base_reg_and_size = arch.get_base_register(reg_offset, size=None)
        if base_reg_and_size is None:
            # give up
            return reg_offset, size
    return base_reg_and_size


@overload
def get_reg_offset_base(
    reg_offset: int, arch: archinfo.Arch, size: int | None = None, resilient: Literal[True] = True
) -> int: ...
@overload
def get_reg_offset_base(
    reg_offset: int, arch: archinfo.Arch, size: int | None = None, resilient: Literal[False] = False
) -> int | None: ...


def get_reg_offset_base(reg_offset, arch, size=None, resilient=True):
    """
    Translate a given register offset into the offset of its full register.

    :param reg_offset:  The offset of the register to translate.
    :param arch:        The corresponding Arch object.
    :param size:        Size of the register to translate. Optional.
    :param resilient:   When set to True, this function will return the provided offset and size for registers that the
                        arch does not know about.
    :return:            The translated offset of the full register.
    """

    base_reg_and_size = arch.get_base_register(reg_offset, size=size)
    if base_reg_and_size is None:
        return reg_offset if resilient else None
    return base_reg_and_size[0]


def get_vvar_deflocs(
    blocks, phi_vvars: dict[int, set[int | None]] | None = None
) -> dict[int, tuple[VirtualVariable, AILCodeLocation]]:
    vvar_to_loc: dict[int, tuple[VirtualVariable, AILCodeLocation]] = {}
    walker = FindExtraDefs()
    walker.found = vvar_to_loc
    for block in blocks:
        for stmt_idx, stmt in enumerate(block.statements):
            if isinstance(stmt, Assignment) and isinstance(stmt.dst, VirtualVariable):
                vvar_to_loc[stmt.dst.varid] = (
                    stmt.dst,
                    AILCodeLocation(block.addr, block.idx, stmt_idx, stmt.tags.get("ins_addr")),
                )
                if phi_vvars is not None and isinstance(stmt.src, Phi):
                    phi_vvars[stmt.dst.varid] = {
                        vvar_.varid if vvar_ is not None else None for src, vvar_ in stmt.src.src_and_vvars
                    }
            elif isinstance(stmt, SideEffectStatement):
                if isinstance(stmt.ret_expr, VirtualVariable):
                    vvar_to_loc[stmt.ret_expr.varid] = (
                        stmt.ret_expr,
                        AILCodeLocation(block.addr, block.idx, stmt_idx, stmt.tags.get("ins_addr")),
                    )
                if isinstance(stmt.fp_ret_expr, VirtualVariable):
                    vvar_to_loc[stmt.fp_ret_expr.varid] = (
                        stmt.fp_ret_expr,
                        AILCodeLocation(block.addr, block.idx, stmt_idx, stmt.tags.get("ins_addr")),
                    )

            if extra_defs := stmt.tags.get("extra_defs", None):
                walker.walk_statement(stmt, block, stmt_idx)
                assert all(varid in vvar_to_loc for varid in extra_defs), "extra_def tag was dropped"

    return vvar_to_loc


def get_vvar_uselocs(blocks) -> dict[int, list[tuple[VirtualVariable, AILCodeLocation]]]:
    vvar_to_loc: dict[int, list[tuple[VirtualVariable, AILCodeLocation]]] = defaultdict(list)
    for block in blocks:
        collector = VVarUsesCollector()
        collector.walk(block)
        for vvar_idx, vvar_and_uselocs in collector.vvar_and_uselocs.items():
            if vvar_idx not in vvar_to_loc:
                vvar_to_loc[vvar_idx] = list(vvar_and_uselocs)
            else:
                vvar_to_loc[vvar_idx] += vvar_and_uselocs
    return vvar_to_loc


def get_tmp_deflocs(blocks: Iterable[Block]) -> dict[Address, dict[atoms.Tmp, int]]:
    tmp_to_loc: dict[Address, dict[atoms.Tmp, int]] = defaultdict(dict)

    for block in blocks:
        codeloc = (block.addr, block.idx)
        for stmt_idx, stmt in enumerate(block.statements):
            if isinstance(stmt, Assignment) and isinstance(stmt.dst, Tmp):
                tmp_to_loc[codeloc][atoms.Tmp(stmt.dst.tmp_idx, stmt.dst.bits)] = stmt_idx
            if isinstance(stmt, CAS):
                if isinstance(stmt.old_lo, Tmp):
                    tmp_to_loc[codeloc][atoms.Tmp(stmt.old_lo.tmp_idx, stmt.old_lo.bits)] = stmt_idx
                if stmt.old_hi is not None and isinstance(stmt.old_hi, Tmp):
                    tmp_to_loc[codeloc][atoms.Tmp(stmt.old_hi.tmp_idx, stmt.old_hi.bits)] = stmt_idx

    return tmp_to_loc


def get_tmp_uselocs(blocks: Iterable[Block]) -> dict[Address, dict[atoms.Tmp, set[tuple[Tmp, int]]]]:
    tmp_to_loc: dict[Address, dict[atoms.Tmp, set[tuple[Tmp, int]]]] = defaultdict(dict)

    for block in blocks:
        collector = TmpUsesCollector()
        collector.walk(block)
        block_loc = (block.addr, block.idx)
        for (tmp_idx, tmp_bits), tmp_and_stmtids in collector.tmp_and_uselocs.items():
            if tmp_idx not in tmp_to_loc[block_loc]:
                tmp_to_loc[block_loc][atoms.Tmp(tmp_idx, tmp_bits)] = tmp_and_stmtids
            else:
                tmp_to_loc[block_loc][atoms.Tmp(tmp_idx, tmp_bits)] |= tmp_and_stmtids

    return tmp_to_loc


def is_const_assignment(stmt: Statement, only_consts: bool = False) -> tuple[bool, Const | StackBaseOffset | None]:
    if isinstance(stmt, Assignment) and (
        isinstance(stmt.src, Const) or (not only_consts and isinstance(stmt.src, StackBaseOffset))
    ):
        return True, stmt.src
    return False, None


class AILBlacklistExprTypeWalker(AILBlockViewer):
    """
    Walks an AIL expression or statement and determines if it does not contain certain types of expressions.
    """

    def __init__(self, blacklist_expr_types: tuple[type, ...], skip_if_contains_vvar: int | None = None):
        super().__init__()
        self.blacklist_expr_types = blacklist_expr_types
        self.has_blacklisted_exprs = False
        self.skip_if_contains_vvar = skip_if_contains_vvar

        self._has_specified_vvar = False

    def _handle_expr(
        self, expr_idx: int, expr: Expression, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> Any:
        if isinstance(expr, self.blacklist_expr_types):
            if self.skip_if_contains_vvar is None:
                self.has_blacklisted_exprs = True
                return None
            # otherwise we do a more complicated check
            self._has_specified_vvar = False  # we do not support nested blacklisted expr types
            has_blacklisted_exprs = True
            r = super()._handle_expr(expr_idx, expr, stmt_idx, stmt, block)
            if self._has_specified_vvar is False:
                # we have seen the vvar that we are looking for! ignore this match
                self.has_blacklisted_exprs = has_blacklisted_exprs
                return None
            self._has_specified_vvar = False
            return r

        return super()._handle_expr(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_VirtualVariable(
        self, expr_idx: int, expr: VirtualVariable, stmt_idx: int, stmt: Statement | None, block: Block | None
    ):
        if self.skip_if_contains_vvar is not None and expr.varid == self.skip_if_contains_vvar:
            self._has_specified_vvar = True
        return super()._handle_VirtualVariable(expr_idx, expr, stmt_idx, stmt, block)


def is_const_and_vvar_assignment(stmt: Statement) -> bool:
    if isinstance(stmt, Assignment):
        walker = AILBlacklistExprTypeWalker((Tmp, Load, Register, Phi, Call, DirtyExpression))
        walker.walk_expression(stmt.src)
        return not walker.has_blacklisted_exprs
    return False


def is_const_vvar_tmp_assignment(stmt: Statement) -> bool:
    if isinstance(stmt, Assignment):
        walker = AILBlacklistExprTypeWalker((Load, Register, Phi, Call, DirtyExpression))
        walker.walk_expression(stmt.src)
        return not walker.has_blacklisted_exprs
    return False


def is_const_vvar_load_assignment(stmt: Statement) -> bool:
    if isinstance(stmt, Assignment):
        walker = AILBlacklistExprTypeWalker((Tmp, Register, Phi, Call, DirtyExpression))
        walker.walk_expression(stmt.src)
        return not walker.has_blacklisted_exprs
    return False


def is_const_vvar_load_dirty_assignment(stmt: Statement) -> bool:
    if isinstance(stmt, Assignment):
        walker = AILBlacklistExprTypeWalker((Tmp, Register, Phi, Call))
        walker.walk_expression(stmt.src)
        return not walker.has_blacklisted_exprs
    return False


def is_phi_assignment(stmt: Statement) -> bool:
    return isinstance(stmt, Assignment) and isinstance(stmt.src, Phi)


def has_load_expr(stmt: Statement, skip_if_contains_vvar: int | None = None) -> bool:
    walker = AILBlacklistExprTypeWalker((Load,), skip_if_contains_vvar=skip_if_contains_vvar)
    walker.walk_statement(stmt)
    return walker.has_blacklisted_exprs


def phi_assignment_get_src(stmt: Statement) -> Phi | None:
    if isinstance(stmt, Assignment) and isinstance(stmt.src, Phi):
        return stmt.src
    return None


def is_dephi_vvar(vvar: VirtualVariable) -> bool:
    return vvar.varid == DEPHI_VVAR_REG_OFFSET


def has_ite_expr(expr: Expression) -> bool:
    walker = AILBlacklistExprTypeWalker((ITE,))
    walker.walk_expression(expr)
    return walker.has_blacklisted_exprs


def has_ite_stmt(stmt: Statement) -> bool:
    walker = AILBlacklistExprTypeWalker((ITE,))
    walker.walk_statement(stmt)
    return walker.has_blacklisted_exprs


def has_tmp_expr(expr: Expression) -> bool:
    walker = AILBlacklistExprTypeWalker((Tmp,))
    walker.walk_expression(expr)
    return walker.has_blacklisted_exprs


class AILReferenceFinder(AILBlockViewer):
    """
    Walks an AIL expression or statement and finds if it contains references to certain expressions.
    """

    def __init__(self, vvar_id: int):
        super().__init__()
        self.vvar_id = vvar_id
        self.has_references_to_vvar = False

    def _handle_UnaryOp(
        self, expr_idx: int, expr: UnaryOp, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> Any:
        if expr.op == "Reference" and isinstance(expr.operand, VirtualVariable) and expr.operand.varid == self.vvar_id:
            self.has_references_to_vvar = True
            return None
        return super()._handle_UnaryOp(expr_idx, expr, stmt_idx, stmt, block)


def has_reference_to_vvar(stmt: Statement, vvar_id: int) -> bool:
    walker = AILReferenceFinder(vvar_id)
    walker.walk_statement(stmt)
    return walker.has_references_to_vvar


def stmt_is_simple_call(stmt: Statement) -> Call | None:
    if isinstance(stmt, SideEffectStatement):
        return stmt.expr if isinstance(stmt.expr, Call) else None
    if not isinstance(stmt, Assignment):
        return None
    src = stmt.src
    while True:
        if isinstance(src, Call):
            return src
        if isinstance(src, Convert):
            src = src.operand
        elif isinstance(src, Extract):
            src = src.base
        elif isinstance(src, Insert):
            src = src.value
        else:
            return None


def check_in_between_stmts(
    graph: networkx.DiGraph,
    blocks: dict[tuple[int, int | None], Block],
    defloc: AILCodeLocation,
    useloc: AILCodeLocation,
    predicate: Callable,
):
    use_block = blocks[(useloc.addr, useloc.block_idx)]
    def_block = blocks[(defloc.addr, defloc.block_idx)]

    # traverse the graph, go from use_block until we reach def_block, and look for Store statements
    seen = {use_block}
    queue = [use_block]
    while queue:
        block = queue.pop(0)

        starting_stmt_idx, ending_stmt_idx = 0, len(block.statements)
        if block is def_block:
            starting_stmt_idx = defloc.stmt_idx + 1
        if block is use_block:
            ending_stmt_idx = useloc.stmt_idx

        for i in range(starting_stmt_idx, ending_stmt_idx):
            if predicate(block.statements[i]):
                return True

        if block is def_block:
            continue

        for pred in graph.predecessors(block):
            if pred not in seen:
                seen.add(pred)
                queue.append(pred)

    return False


def has_store_stmt_in_between_stmts(
    graph: networkx.DiGraph,
    blocks: dict[tuple[int, int | None], Block],
    defloc: AILCodeLocation,
    useloc: AILCodeLocation,
) -> bool:
    return check_in_between_stmts(graph, blocks, defloc, useloc, lambda stmt: isinstance(stmt, Store))


def has_call_in_between_stmts(
    graph: networkx.DiGraph,
    blocks: dict[tuple[int, int | None], Block],
    defloc: AILCodeLocation,
    useloc: AILCodeLocation,
    skip_if_contains_vvar: int | None = None,
) -> bool:
    def _contains_call(stmt: Statement) -> bool:
        if isinstance(stmt, SideEffectStatement):
            return True
        # walk the statement and check if there is a call expression
        walker = AILBlacklistExprTypeWalker((Call,), skip_if_contains_vvar=skip_if_contains_vvar)
        walker.walk_statement(stmt)
        return walker.has_blacklisted_exprs

    return check_in_between_stmts(graph, blocks, defloc, useloc, _contains_call)


def has_load_expr_in_between_stmts(
    graph: networkx.DiGraph,
    blocks: dict[tuple[int, int | None], Block],
    defloc: AILCodeLocation,
    useloc: AILCodeLocation,
    skip_if_contains_vvar: int | None = None,
) -> bool:
    return check_in_between_stmts(
        graph, blocks, defloc, useloc, lambda stmt: has_load_expr(stmt, skip_if_contains_vvar=skip_if_contains_vvar)
    )


def is_vvar_propagatable(vvar: VirtualVariable, def_stmt: Statement | None, stack_arg_offsets: set[int] | None) -> bool:
    if isinstance(def_stmt, Assignment) and isinstance(def_stmt.src, Insert):
        # do not create huge insert chains
        return False
    if vvar.was_tmp or vvar.was_reg or vvar.was_parameter:
        return True
    if vvar.was_stack and isinstance(def_stmt, Assignment):
        if (
            stack_arg_offsets is not None
            and vvar.stack_offset in stack_arg_offsets
            and not isinstance(def_stmt.src, Phi)
        ):
            return True
        if (
            isinstance(def_stmt.src, VirtualVariable)
            and def_stmt.src.was_stack
            and def_stmt.src.stack_offset == vvar.stack_offset
        ):
            # special case: the following block
            #   ## Block 401e98
            #   00 | 0x401e98 | LABEL_401e98:
            #   01 | 0x401e98 | vvar_227{stack -12} = 𝜙@32b [((4202088, None), vvar_277{stack -12}), ((4202076, None),
            #                   vvar_278{stack -12})]
            #   02 | 0x401ea0 | return Conv(32->64, vvar_227{stack -12});
            # might be simplified to the following block after return duplication
            #   ## Block 401e98.1
            #   00 | 0x401e98 | LABEL_401e98__1:
            #   01 | 0x401e98 | vvar_279{stack -12} = vvar_277{stack -12}
            #   02 | 0x401ea0 | return Conv(32->64, vvar_279{stack -12});
            # in this case, vvar_279 is eliminatable.
            return True
    return False


def is_vvar_eliminatable(vvar: VirtualVariable, def_stmt: Statement | None) -> bool:
    if vvar.was_tmp or vvar.was_reg or vvar.was_parameter:
        return True
    if (  # noqa: SIM103
        vvar.was_stack
        and isinstance(def_stmt, Assignment)
        and isinstance(def_stmt.src, VirtualVariable)
        and def_stmt.src.was_stack
        and def_stmt.src.stack_offset == vvar.stack_offset
    ):
        # special case: the following block
        #   ## Block 401e98
        #   00 | 0x401e98 | LABEL_401e98:
        #   01 | 0x401e98 | vvar_227{stack -12} = 𝜙@32b [((4202088, None), vvar_277{stack -12}), ((4202076, None),
        #                   vvar_278{stack -12})]
        #   02 | 0x401ea0 | return Conv(32->64, vvar_227{stack -12});
        # might be simplified to the following block after return duplication
        #   ## Block 401e98.1
        #   00 | 0x401e98 | LABEL_401e98__1:
        #   01 | 0x401e98 | vvar_279{stack -12} = vvar_277{stack -12}
        #   02 | 0x401ea0 | return Conv(32->64, vvar_279{stack -12});
        # in this case, vvar_279 is eliminatable.
        return True
    return False


__all__ = (
    "VVarUsesCollector",
    "check_in_between_stmts",
    "get_tmp_deflocs",
    "get_tmp_uselocs",
    "get_vvar_deflocs",
    "get_vvar_uselocs",
    "has_call_in_between_stmts",
    "has_ite_expr",
    "has_load_expr_in_between_stmts",
    "has_store_stmt_in_between_stmts",
    "is_const_and_vvar_assignment",
    "is_const_assignment",
    "is_const_vvar_load_assignment",
    "is_const_vvar_load_dirty_assignment",
    "is_phi_assignment",
    "is_vvar_eliminatable",
    "phi_assignment_get_src",
)



from __future__ import annotations
import random
import string

from angr.analyses.identifier.func import Func, TestData

BASES = [8, 10, 16]


digs = string.digits + string.ascii_letters


def int2base(x, base):
    if x < 0:
        sign = -1
    elif x == 0:
        return digs[0]
    else:
        sign = 1
    x *= sign
    digits = []
    while x:
        digits.append(digs[x % base])
        x //= base
    if sign < 0:
        digits.append("-")
    digits.reverse()
    return "".join(digits)


class OneTwoOrThree:
    def __eq__(self, other):
        if other == 1:
            return True
        if other == 2:
            return True
        return other == 3

    def __ne__(self, other):
        if other == 1:
            return False
        if other == 2:
            return False
        return not other == 3


class based_atoi(Func):
    def __init__(self):
        super().__init__()
        self.allows_negative = True
        self.base = None

    def rand_str(self, length, byte_list=None):  # pylint disable=no-self-use
        if byte_list is None:
            return "".join(chr(random.randint(0, 255)) for _ in range(length))
        return "".join(random.choice(byte_list) for _ in range(length))

    def num_args(self):
        return OneTwoOrThree()

    def args(self):  # pylint disable=no-self-use
        return ["buf", "size", "err"]

    def get_name(self):
        name = "based_atoi_"
        if self.allows_negative:
            name += "signed_"
        name += str(self.base)
        return name

    def gen_input_output_pair(self):
        num = random.randint(-(2**26), 2**26 - 1)

        if not self.allows_negative:
            num = abs(num)

        num = abs(num)
        s = int2base(num, self.base)
        test_input = [s, 30, b"foo"]
        test_output = [s, None, None]
        return_val = num
        max_steps = 10
        return TestData(test_input, test_output, return_val, max_steps)

    def pre_test(self, func, runner):
        s = "1111"
        test_input = [s, 10, "foo"]
        max_steps = 10
        test = TestData(test_input, [None, None, None], 1234, max_steps=max_steps)
        state = runner.get_out_state(func, test, concrete_rand=True)
        if state is None:
            return False
        out_val = state.solver.eval(state.regs.eax)
        self.base = None
        for i in range(2, 16):
            if out_val == int(s, i):
                self.base = i
                break
        if self.base is None:
            return False

        num = random.randint(-(2**26), 2**26 - 1)

        num = abs(num)
        s = int2base(num, self.base)
        test_input = [s, 30, "foo"]
        test_output = [s, None, None]
        return_val = num
        max_steps = 10
        test = TestData(test_input, test_output, return_val, max_steps)
        if not runner.test(func, test):
            return False

        num = -random.randint(2000, 8000)
        s = int2base(num, self.base)
        test_input = [s, 30, "foo"]
        test_output = [s, None, None]
        return_val = num
        max_steps = 10
        test = TestData(test_input, test_output, return_val, max_steps)
        if not runner.test(func, test):
            self.allows_negative = False

        return True



from __future__ import annotations
import angr.ailment as ailment


class CCallRewriterBase:
    """
    The base class for CCall rewriters.
    """

    __slots__ = (
        "project",
        "result",
    )

    def __init__(self, ccall: ailment.Expr.VEXCCallExpression, project, rename_ccalls: bool = False):
        self.project = project
        self.result: ailment.Expr.Expression | None = self._rewrite(ccall)
        if rename_ccalls and self.result is None and ccall.callee != "_ccall":
            renamed = ccall.copy()
            renamed.callee = "_ccall"
            self.result = renamed

    def _rewrite(self, ccall: ailment.Expr.VEXCCallExpression) -> ailment.Expr.Expression | None:
        raise NotImplementedError



from __future__ import annotations
import angr


class getsockopt(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, sockfd, level, optname, optval, optlen):
        # TODO: ...

        return 0



from __future__ import annotations

import claripy

from .base import SimConcretizationStrategy


class SimConcretizationStrategySignedAdd(SimConcretizationStrategy):
    """
    Concretization strategy that changes additions of big offsets to subtractions of small offsets.
    """

    def __init__(self, subtraction_limit=0x10000):
        super().__init__()
        self._subtraction_limit = subtraction_limit

    def _concretize(self, memory, addr, **kwargs):
        if addr.depth == 2 and addr.op == "__add__":
            if addr.args[0].singlevalued and addr.args[1].symbolic:
                # Swap variable and immediate
                addr.args = (addr.args[1], addr.args[0])
            if (
                addr.args[0].symbolic
                and addr.args[1].singlevalued
                # Check if negative argument
                and memory.state.solver.is_true(addr.args[1] >= 1 << (addr.args[1].size() - 1))
            ):
                new_arg = (1 << addr.args[1].size()) - memory.state.solver.eval(addr.args[1])
                if new_arg < self._subtraction_limit:
                    addr.op = "__sub__"
                    addr.args = (addr.args[0], claripy.BVV(new_arg, addr.args[1].size()))



# pylint:disable=missing-class-docstring,missing-function-docstring
"""
Semantic variable naming for function call results.

This module detects variables that store function call results and names them
based on the called function (e.g., malloc result -> ptr, strlen result -> len).
"""

from __future__ import annotations
from typing import TYPE_CHECKING
import logging

from angr import ailment
from angr.ailment.expression import Call
from angr.ailment.statement import Assignment, SideEffectStatement
from angr.sim_variable import SimVariable

from .naming_base import ClinicNamingBase

if TYPE_CHECKING:
    pass

l = logging.getLogger(name=__name__)

# Mapping of function names/patterns to suggested variable names
FUNCTION_RESULT_NAMES = {
    # Memory allocation
    "malloc": "ptr",
    "calloc": "ptr",
    "realloc": "ptr",
    "alloca": "ptr",
    "mmap": "ptr",
    "new": "ptr",
    # String functions
    "strlen": "len",
    "wcslen": "len",
    "strnlen": "len",
    "strcpy": "dst",
    "strncpy": "dst",
    "strcat": "dst",
    "strncat": "dst",
    "strchr": "ptr",
    "strrchr": "ptr",
    "strstr": "ptr",
    "strdup": "str",
    "strtok": "tok",
    # Memory functions
    "memcpy": "dst",
    "memmove": "dst",
    "memset": "dst",
    "memchr": "ptr",
    # File I/O
    "fopen": "fp",
    "fdopen": "fp",
    "freopen": "fp",
    "tmpfile": "fp",
    "fread": "count",
    "fwrite": "count",
    "fgets": "str",
    "fgetc": "ch",
    "getc": "ch",
    "getchar": "ch",
    # Process/system
    "fork": "pid",
    "getpid": "pid",
    "getppid": "pid",
    "getuid": "uid",
    "geteuid": "uid",
    "getgid": "gid",
    "getegid": "gid",
    # Network
    "socket": "sock",
    "accept": "sock",
    "recv": "count",
    "recvfrom": "count",
    "send": "count",
    "sendto": "count",
    "read": "count",
    "write": "count",
    # Error handling
    "errno": "err",
    "strerror": "msg",
    "perror": "msg",
    # Conversion
    "atoi": "num",
    "atol": "num",
    "atoll": "num",
    "strtol": "num",
    "strtoul": "num",
    "strtoll": "num",
    "strtod": "num",
    # Time
    "time": "t",
    "clock": "t",
    "localtime": "tm",
    "gmtime": "tm",
    # Other common functions
    "sizeof": "size",
    "count": "count",
    "size": "size",
    "length": "len",
}


class CallResultNaming(ClinicNamingBase):
    """
    Detects variables that store function call results and names them
    based on the called function.
    """

    PRIORITY = 40  # Before pointer naming, since call results are more specific

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._result_vars: dict[SimVariable, str] = {}  # var -> suggested name

    def analyze(self) -> dict[SimVariable, str]:
        """
        Analyze the graph for function call result patterns.

        :return: Dictionary mapping SimVariable to new name
        """
        if not self._graph:
            return {}

        # Find all variables that receive function call results
        self._find_call_results()

        # Convert to var_to_new_name
        self._var_to_new_name = dict(self._result_vars)

        return self._var_to_new_name

    def _find_call_results(self) -> None:
        """
        Find variables that store function call results.
        """
        for node in self._graph:
            if not isinstance(node, ailment.Block):
                continue

            for stmt in node.statements:
                # Look for Call statements with return values
                if isinstance(stmt, SideEffectStatement):
                    self._analyze_call(stmt)

                # Look for assignments where src is a call
                elif isinstance(stmt, Assignment) and isinstance(stmt.src, Call):
                    self._analyze_call_assignment(stmt)

    def _analyze_call(self, call: SideEffectStatement) -> None:
        """
        Analyze a Call statement for its return value.
        """
        if call.ret_expr is None:
            return

        # Get the variable storing the result
        var = self._get_linked_variable(call.ret_expr)
        if var is None:
            return

        # Get the function name
        func_name = self._get_function_name(call.expr)
        if func_name is None:
            return

        # Check if we have a naming rule for this function
        suggested_name = self._get_name_for_function(func_name)
        if suggested_name:
            self._record_result_var(var, suggested_name)

    def _analyze_call_assignment(self, stmt: Assignment) -> None:
        """
        Analyze an assignment where the source is a function call.
        """
        call = stmt.src
        if not isinstance(call, Call):
            return

        # Get the variable being assigned to
        var = self._get_linked_variable(stmt.dst)
        if var is None:
            return

        # Get the function name
        func_name = self._get_function_name(call)
        if func_name is None:
            return

        # Check if we have a naming rule for this function
        suggested_name = self._get_name_for_function(func_name)
        if suggested_name:
            self._record_result_var(var, suggested_name)

    @staticmethod
    def _get_name_for_function(func_name: str) -> str | None:
        """
        Get the suggested variable name for a function's result.
        """

        # Normalize the function name
        normalized = func_name.lower().strip("_")

        # Check for exact match
        if normalized in FUNCTION_RESULT_NAMES:
            return FUNCTION_RESULT_NAMES[normalized]

        # Check for partial matches (e.g., __malloc -> malloc)
        for pattern, name in FUNCTION_RESULT_NAMES.items():
            if pattern in normalized:
                return name

        return None

    def _record_result_var(self, var: SimVariable, suggested_name: str) -> None:
        """
        Record a variable as a function result.
        """
        # Don't overwrite if already recorded with a name
        if var not in self._result_vars:
            self._result_vars[var] = suggested_name
